# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 04.09.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
A variety of functions used to preprocess an image for GLCM analysis.
"""

# pylint: disable=bad-indentation

import numpy as np
import cv2

def scale_image(image, min_val=0, max_val=255):
  """Scale values of input image to range [min_val, max_val]"""
  im_max = np.max(image)
  im_min = np.min(image)
  return min_val + (1 - (im_max - image)/(im_max - im_min))*(max_val - min_val)

def crop_image(full_image, image_part='full_image'):
  """Crop input image. Usefull if you want to test out on smaller image
  chunks.

  Args:
    full_image: 2D numpy array, shape [M, N]. Image to be cropped.

  Returns:
    image: 2D numpy array, 8bit valued, shape [M, N]. Result image.
  """
  M, N = full_image.shape

  if image_part is 'full_image':
    image = full_image
  elif image_part is 'top_left':
    image = full_image[:M/2, :N/2]
  elif image_part is 'top_right':
    image = full_image[:M/2, N/2:]
  elif image_part is 'bottom_left':
    image = full_image[M/2:, :N/2]
  elif image_part is 'bottom_right':
    image = full_image[M/2:, N/2:]
  elif image_part is 'custom_image':
    image = full_image[0:30, 0:30]
  else:
    print(image_part, ' is not a valid image segment, set to full.')
    image = full_image

  return image

def resize_image(full_image, new_height, new_width):
  """Resize full image to shape [new_height, new_width]"""
  if new_height is None and new_width is None:
    return full_image
  else:
    if new_height is None:
      new_height = full_image.shape[0]
    if new_width is None:
      new_width = full_image.shape[1]
    return cv2.resize(full_image, (new_height, new_width),
                      interpolation=cv2.INTER_AREA)


def histogram_equalization(image):
  """Perform histogram equalization on input image

  Args:
    image: 2D numpy array, 8bit valued, shape [M, N]. Image to be equalized.

  Returns:
    out_image: 2D numpy array, 8bit valued, shape [M, N]. Result image.
  """

  M, N = image.shape

  # Normalized histogram of original image, which we can interpret as a
  # probability mass function of the gray-level intensities in the image.
  histogram, _ = np.histogram(image.flatten(), bins=256, range=[0, 256], density=True)

  # Continuing this interpretation, we can construct a cumulative mass function
  # from this histogram.
  cumulative = np.cumsum(histogram)

  # Use this to equalize the image.
  out_image = np.zeros([M, N])
  for i in range(M):
    for j in range(N):
      out_image[i, j] = np.floor(255. * cumulative[image[i, j]])

  return out_image


def requantize(image, num_levels=None):
  """Requantize input image by creating a list of values, and for every pixel
  in the image, set its new value equal to the closest value in this list.

  Args:
    image: 2D numpy array, 8bit valued, shape [M, N]. Image to be requantized.
    num_levels: int. Number of intensity values in the output image.

  Returns:
    out_image: uint8 2D numpy array, 8bit valued, shape [M, N]. Result image.
  """
  if num_levels is None:
    print('Please provide number of levels in requantization')
    return image

  M, N = image.shape

  out_image = np.zeros([M, N])
  levels = np.linspace(0, 255, num_levels)
  for i in range(M):
    for j in range(N):
      out_image[i, j] = min(levels, key=lambda x, ii=i, jj=j: abs(x - image[ii, jj]))

  return out_image.astype('uint8')

def preprocess(image, num_greylevels, new_height=None, new_width=None):
  """Preprocess image using a variety of functions

  Args:
    image: 2D numpy array, shape [M, N]. Image to be preprocessed.
    num_greylevels: int. Number of graylevels in image after requantization.
    new_height: int. Height of output image.
    new_width: int. Width of output image.

  Returns:
    image: 2D numpy array, 8bit valued, shape [M, N]. Result image.
  """

  #image = crop_image(image, image_part=image_part)
  image = resize_image(image, new_height, new_width)
  image = histogram_equalization(image)
  image = requantize(image, num_levels=num_greylevels)

  return image
